#!/usr/bin/env ruby
# -*- coding: utf-8; mode: ruby -*-
#
# inplace_mini - edits files in-place through a given filter command
#
# Copyright (c) 2004, 2005, 2006, 2007, 2008 Akinori MUSHA
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

MYNAME = File.basename($0)

require "optparse"

def main(argv)
  usage = <<-"EOF"
usage: #{MYNAME} [-Lfitz] [-b SUFFIX] COMMANDLINE [file ...]
  EOF

  banner = <<-"EOF"
#{MYNAME} - edits files in-place through a given filter command

#{usage}
  EOF

  opts = OptionParser.new(banner, 24) { |opts|
    opts.on("-h", "--help",
      "Show this message.") {
      print opts
      exit 0
    }

    opts.on("-L", "--dereference",
      "Edit the original file for each symlink.") {
      |b| $dereference = b
    }

    opts.on("-b", "--backup-suffix=SUFFIX",
      "Create a backup file with the SUFFIX for each file.") {
      |s| $backup_suffix = s
    }

    opts.on("-f", "--force",
      "Force editing even if a file is read-only.") {
      |b| $force = b
    }

    opts.on("-i", "--preserve-inode",
      "Make sure to preserve the inode number of each file.") {
      |b| $preserve_inode = b
    }

    opts.on("-t", "--preserve-timestamp",
      "Preserve the modification time of each file.") {
      |b| $preserve_time = b
    }

    opts.on("-z", "--accept-empty",
      "Accept empty (zero-sized) output.") {
      |b| $accept_empty = b
    }
  }
  # $B=i4|2=(B
  setup()
  # $B%3%^%s%I%i%$%s$r%Q!<%9(B
  template, *files = opts.order(*argv)
  # $B:G=i$N0z?t$,%3%^%s%I%i%$%s%F%s%W%l!<%H(B
  if template.nil?
    STDERR.puts "No filter command line to execute given.", ""
    print opts
    exit 1
  end
  # $B;D$j$N0z?t$,=hM}BP>]%U%!%$%k72(B
  if files.empty?
    STDERR.puts "No files to process given.", ""
    print opts
    exit 2
  end
  # $B%U%#%k%?$r@8@.(B
  filter = FileFilter.new(template)
  # $B%U%!%$%k$r=g$K=hM}(B
  files.each { |file|
    begin
      filter.filter!(file, file)
    rescue => e
      STDERR.puts "#{file}: skipped: #{e}"
    end
  }
rescue OptionParser::ParseError => e
  STDERR.puts "#{MYNAME}: #{e}", usage
  exit 64
rescue => e
  STDERR.puts "#{MYNAME}: #{e}"
  exit 1
end

def setup
  $backup_suffix = nil
  $dereference = $force =
    $preserve_inode = $preserve_time = $accept_empty = false
end

require 'set'
require 'tempfile'
require 'fileutils'
require 'pathname'

class FileFilter
  def initialize(template)
    @formatter = Formatter.new(template)
  end

  def destructive?
    @formatter.arity == 1
  end

  def filter!(origfile, file)
    filter(origfile, file, file)
  end

  def filter(origfile, infile, outfile)
    if !File.exist?(infile)
      flunk origfile, "file not found"
    end
    # outfile$B$,0l;~%U%!%$%k$G$J$$!a85%U%!%$%k(B
    outfile_is_original = !tmpfile?(outfile)
    outfile_stat = File.lstat(outfile)
    # $B%7%s%\%j%C%/%j%s%/$N>l9g$r9MN8(B
    if outfile_stat.symlink?
      # $dereference$B$,??$G$J$$>l9g$O%9%-%C%W(B
      $dereference or
        flunk origfile, "symlink"
      # $B%7%s%\%j%C%/%j%s%/$rE83+!J2r7h!K$9$k(B
      begin
        outfile = Pathname.new(outfile).realpath.to_s
        outfile_stat = File.lstat(outfile)
      rescue => e
        flunk origfile, "symlink unresolvable: %s", e
      end
    end
    # $BDL>o%U%!%$%k0J30$O07$o$J$$(B
    outfile_stat.file? or
      flunk origfile, "symlink to a non-regular file"
    # $force$B$,??$N$H$-0J30$OFI<h@lMQ%U%!%$%k$O$$$8$i$J$$(B
    $force || outfile_stat.writable? or
      flunk origfile, "symlink to a read-only file"
    # $B0l;~%U%!%$%k$r:n@.(B
    tmpfile = FileFilter.make_tmpfile_for(outfile)
    # $B%U%#%k%?%3%^%s%I$OGK2uE*$+!)(B
    if destructive?
      # $B0l;~%U%!%$%k$K%3%T!<$7$F$=$l$r=hM}$5$;$k(B
      FileUtils.cp(infile, tmpfile)
      command = @formatter.format(origfile, tmpfile)
    else
      command = @formatter.format(origfile, infile, tmpfile)
    end
    # $B%U%#%k%?%3%^%s%I$r<B9T(B
    if system(command)
      File.file?(tmpfile) or
        flunk origfile, "output file removed"
      # $B7k2L$NCf?H$,6u$N>l9g(B
      !$accept_empty && File.zero?(tmpfile) and
        flunk origfile, "empty output"
      # $BCf?H$,JQ$o$C$F$$$J$$>l9g(B
      outfile_is_original &&
       FileUtils.identical?(origfile, tmpfile) and
        flunk origfile, "unchanged"
      # $B85%U%!%$%k$NB0@-$r<hF@(B
      stat = File.stat(infile)
      # $B%7%0%J%k$K$h$k=*N;$r6X;_$7$FCV$-49$(=hM}(B
      uninterruptible {
        replace(tmpfile, outfile, stat)
      }
      # $B%5%$%:$NJQ2=$rJs9p(B
      info "%s: edited (%d bytes -> %d bytes)",
           origfile, stat.size, File.size(outfile)
    else
      flunk origfile, "command exited with %d", $?.exitstatus
    end
  end

  @@tmpfiles = Set.new

  # $B0l;~%U%!%$%k$+$I$&$+$NH=Dj(B
  def tmpfile?(file)
    @@tmpfiles.include?(file)
  end

  # $B;XDj%U%!%$%k$HF1$83HD%;R$N0l;~%U%!%$%k$r:n@.(B
  def self.make_tmpfile_for(outfile)
    if m = File.basename(outfile).match(/(\..+)$/)
      ext = m[1]
    else
      ext = ''
    end
    # $BBh0l0z?t$K(B [prefix, suffix] $B$NG[Ns$rEO$9(B
    tmpf = Tempfile.new(['inplace', ext])
    # $B$9$0JD$8$F%Q%9L>$rF@$k(B
    tmpf.close
    path = tmpf.path
    # $B0l;~%U%!%$%k%j%9%H$KDI2C(B
    @@tmpfiles << path
    return path
  end

  private
  def info(fmt, *args)
    puts sprintf(fmt, *args)
  end

  def error(fmt, *args)
    STDERR.puts "error: " + sprintf(fmt, *args)
  end

  def flunk(origfile, fmt, *args)
    raise "#{origfile}: " << sprintf(fmt, *args)
  end

  def replace(file1, file2, stat)
    if tmpfile?(file2)
      # $B0l;~%U%!%$%k$J$i(Bi-node$B$d%P%C%/%"%C%W$O9MN8ITMW(B
      FileUtils.mv(file1, file2)
    else
      # file2$B$,0l;~%U%!%$%k$G$O$J$$!a85%U%!%$%k(B
      if $backup_suffix && !$backup_suffix.empty?
        # $B;XDj$,$"$l$P%P%C%/%"%C%W%U%!%$%k$r:n$k(B
        bakfile = file2 + $backup_suffix
        if $preserve_inode
          # i-node$B$r0];}$7$?$$$J$i%3%T!<(B
          FileUtils.cp(file2, bakfile, :preserve => true)
        else
          # i-node$B$r0];}$7$J$/$F$h$$$J$i%j%M!<%`(B
          FileUtils.mv(file2, bakfile)
        end
      end

      begin
        if $preserve_inode
          # i-node$B$r0];}$7$?$$$J$i>e=q$-%3%T!<(B
          FileUtils.cp(file1, file2)
          FileUtils.rm(file1)
        else
          # i-node$B$r0];}$7$J$/$F$h$$$J$i%j%M!<%`(B
          FileUtils.mv(file1, file2)
        end
      rescue => e
        # $B85%U%!%$%k$X$N=q$-9~$_$N<:GT$OCWL?E*(B
        error "%s: failed to overwrite: %s", file2, e
        error "%s: result file left: %s", file2, file1
        # $B0l;~%U%!%$%k$r;D$7$FB(;~=*N;(B
        exit! 1
      end
    end
    # $B%U%!%$%kB0@-$rI|85(B
    preserve(file2, stat)
  end

  def preserve(file, stat)
    # stat$B$O(BFile::Stat$B%*%V%8%'%/%H(B
    if $preserve_time
      # $preserve_time$B;XDj;~$N$_(B
      File.utime stat.atime, stat.mtime, file
    end

    mode = stat.mode

    begin
      File.chown stat.uid, stat.gid, file
    rescue Errno::EPERM
      # chown$B$G$-$J$$>l9g$O(Bsetuid$B%S%C%HEy$OD|$a$k(B
      mode &= 01777
    end

    File.chmod mode, file
  end

  class Formatter
    def initialize(template)
      @template = template.dup  # $BJ8;zNs$rJ#@=$7$FJ];}(B

      # $B<B:]$K8F$S=P$7$F$_$F%A%'%C%/!J"$!K(B
      self.format("0", "1", "2")

      if @arity == 0
        @template = "(#{@template}) < %1 > %2"
        @arity = 2
      end
    end

    attr_reader :template, :arity

    def format(origfile, infile, outfile = nil)
      s = ''            # $BJV$jCM$NJ8;zNs$rF~$l$k4o(B
      template = @template.dup
                        # $B:n6HMQ%U%)!<%^%C%HJ8;zNs(B
      arity_bits = 0    # %1, %2$B$NM-L5$N5-O?(B

      until template.empty?
        # %$B$,Mh$k$^$G$=$N$^$^%3%T!<(B
        template.sub!(/\A([^%]+)/) {
          s << $1
          ''
        }
        # %0, %1, %2, %%$B$r$=$l$>$lCV49$7$F%3%T!<(B
        template.sub!(/\A%(.)/) {
          case c = $1
          when '%'
            s << c
          when '0'
            s << origfile.shellescape
          when '1'
            s << infile.shellescape
            arity_bits |= 0x1
          when '2'
            s << outfile.shellescape
            arity_bits |= 0x2
          else
            # %$B$N%(%9%1!<%W$7K:$l$O%(%i!<(B
            raise ArgumentError, "invalid placeholder specification (%#{c}): #{@template}"
          end
          ''
        }
      end
      # $B2?$H2?$,;XDj$5$l$?$+$r8+$k(B
      case arity_bits
      when 0x0
        @arity = 0
      when 0x1
        @arity = 1
      when 0x2
        raise ArgumentError, "%1 is missing while %2 is specified: #{@template}"
      when 0x3
        @arity = 2
      end
      # $B@07A$5$l$?J8;zNs$rJV$9(B
      return s
    end
  end
end

if RUBY_VERSION >= "1.8.7"
  require 'shellwords'
else
  class String
    def shellescape
      # An empty argument will be skipped, so return empty quotes.
      return "''" if empty?

      str = dup

      # Process as a single byte sequence because not all shell
      # implementations are multibyte aware.
      str.gsub!(/([^A-Za-z0-9_\-.,:\/@\n])/n, "\\\\\\1")

      # A LF cannot be escaped with a backslash because a backslash + LF
      # combo is regarded as line continuation and simply ignored.
      str.gsub!(/\n/, "'\n'")

      return str
    end
  end

  class Tempfile
    alias orig_make_tmpname make_tmpname

    def make_tmpname(basename, n)
      case basename
      when Array
        prefix, suffix = *basename
        make_tmpname(prefix, n) + suffix
      else
        orig_make_tmpname(basename, n).tr('.', '-')
      end
    end
  end
end

$uninterruptible = false

[:SIGINT, :SIGQUIT, :SIGTERM].each { |sig|
  trap(sig) {
    unless $uninterruptible
      STDERR.puts "Interrupted."
      exit 130  # $BCfCG;~$O(B130$B$rJV$9$N$,:nK!(B
    end
  }
}

def uninterruptible
  # $B85$NCM$rJ]B8$7$F%U%i%0$rN)$F$k(B
  orig = $uninterruptible
  $uninterruptible = true
  yield     # $B%V%m%C%/$r8F$S=P$9(B
ensure
  # $B85$NCM$rI|85$9$k(B
  $uninterruptible = orig
end

main(ARGV)
